
#include "DelphiShared.h"

DelPhiShared::DelPhiShared()
{
	epsmap = NULL;
	buildEpsMap = false;
	buildStatus = false;
	x = NULL;
	y = NULL;
	z = NULL;
	idebmap = NULL;
	atoms = NULL;
	scspos = NULL;
	scsnor = NULL;
	scsarea = NULL;
	ibgp = NULL;
	status = NULL;
	cavitiesVec = NULL;
	delphiBinding = false;
}

void DelPhiShared::finalizeBinding()
{
	// update indexes to be conformant to Fortran
	for (int i=0;i<nbgp;i++)
	{
	  ibgp[3*i]++;
	  ibgp[3*i+1]++;
	  ibgp[3*i+2]++;
	}

	epsmap = NULL; //avoid deletion
	scspos = NULL; //avoid deletion
	scsnor = NULL; //avoid deletion
	idebmap = NULL; //avoid deletion
	//scsarea = NULL; //for now not used, so it must be distroied
	ibgp = NULL;   //avoid deletion	
}


bool DelPhiShared::loadAtoms(char fn[BUFLEN])
{
	// load atoms file
    ifstream fin;
    fin.open(fn,ios::in);
	char buffer[BUFLEN];

    if (fin.fail())
	{
		cout << endl << WARN << "Cannot read file " << fn;
		return false;
	}

	vector<Atom*> list;
	Atom* atm;

    while (!fin.eof( ))  
    {
        fin.getline(buffer,BUFLEN);
		if (strlen(buffer)>1)
		{
			atm = new Atom();
			sscanf(buffer,"%lf %lf %lf %lf",&(atm->pos[0]),&(atm->pos[1]),&(atm->pos[2]),&(atm->radius));               
			// debug
			//printf("\n %lf %lf %lf %lf",atm->pos[0],atm->pos[1],atm->pos[2],atm->radius);               
			list.push_back(atm);
		}
		else
		{
			// skip empty line
		}
    }
    fin.close();    
	
	// freeze list
	vector<Atom*>::iterator it = list.begin();
	if (atoms!=NULL)
	{
		for (int i=0;i<numAtoms;i++)
			delete atoms[i];
	}

	atoms = new Atom*[list.size()];
	
	for (int i=0;it!=list.end();it++,i++)
	{
		atoms[i] = (*it);
	}

	numAtoms = list.size();
	cout << endl << INFO << "Read " << numAtoms << " atoms";
	list.clear();

	if (numAtoms<4)
	{
		cout << endl << ERR << "NanoShaper needs at least 4 atoms to work.";
		cout << endl << REMARK << "To emulate 4 atoms you can place dummy atoms with null radius";
		cout << endl << REMARK << "at the center of the real atoms";
		exit(-1);
	}

	return true;

}

bool DelPhiShared::loadAtoms(int na,double* x,double* r,double* q,int* d)
{
	cout << endl << INFO << "Read " << na << " atoms";
	if (atoms!=NULL)
	{
		for (int i=0;i<numAtoms;i++)
			delete atoms[i];
	}

	if (r==NULL || x==NULL)
	{
		cout << endl << WARN << "Atoms info not available!";
		return false;
	}

	atoms = new Atom*[na];
	numAtoms = na;
	
	for (int i=0;i<na;i++)
	{
		double qq;
		int dd;
		if (q==NULL)
			qq = 0;
		else
			qq=q[i];

		if (d==NULL)
			dd = 0;
		else
			dd=d[i];

		atoms[i] = new Atom(x[i*3],x[i*3+1],x[i*3+2],r[i],qq,dd);		
	}

	return true;
}
	
bool DelPhiShared::buildGrid(double scale,double perfill)
{
	if (atoms==NULL)
	{
		cout << endl << ERR << "Cannot build grid with no atoms!";
		return false;
	}

	if (scale<0)
	{
		cout << endl << WARN << "Cannot use a <0 scale: setting " << DEFAULT_SCALE;
		scale = DEFAULT_SCALE;
	}

	if (perfill < 0 || perfill > 100)
	{
		cout << endl << WARN << "Perfil is in (0,100). Setting " << DEFAULT_PERFIL;
		scale = DEFAULT_PERFIL;	
	}

	this->scale = scale;
	this->perfill = perfill;

	double cmin[3] = {INFINITY,INFINITY,INFINITY};
	double cmax[3] = {-INFINITY,-INFINITY,-INFINITY};
	double oldmid[3];

	for (int ia=0;ia<numAtoms;ia++)
	{
		cmin[0] = min(cmin[0],atoms[ia]->pos[0]-atoms[ia]->radius);		
		cmin[1] = min(cmin[1],atoms[ia]->pos[1]-atoms[ia]->radius);
		cmin[2] = min(cmin[2],atoms[ia]->pos[2]-atoms[ia]->radius);

		cmax[0] = max(cmax[0],atoms[ia]->pos[0]+atoms[ia]->radius);		
		cmax[1] = max(cmax[1],atoms[ia]->pos[1]+atoms[ia]->radius);
		cmax[2] = max(cmax[2],atoms[ia]->pos[2]+atoms[ia]->radius);
	}

	oldmid[0]=(cmax[0]+cmin[0])/2.;
	oldmid[1]=(cmax[1]+cmin[1])/2.;
	oldmid[2]=(cmax[2]+cmin[2])/2.;

	cout << endl << INFO << "Geometric baricenter ->  " << oldmid[0] << 
		" " << oldmid[1] << " " << oldmid[2];

	double v[6];
	v[0] = fabs(cmax[0]-oldmid[0]);
	v[1] = fabs(cmin[0]-oldmid[0]);
	v[2] = fabs(cmax[1]-oldmid[1]);
	v[3] = fabs(cmin[1]-oldmid[1]);
	v[4] = fabs(cmax[2]-oldmid[2]);
	v[5] = fabs(cmin[2]-oldmid[2]);

	rmaxdim = -INFINITY;

	for (int i=0;i<6;i++)
		if (rmaxdim<v[i])
			rmaxdim = v[i];

	rmaxdim = 2*rmaxdim;

	       
	unsigned int igrid = (unsigned int)floor(scale*100./perfill*rmaxdim);

	if ((igrid%2)==0)
		igrid++;
	
	cout << endl << INFO << "Grid is " << igrid;

	baricenter[0] = oldmid[0];
	baricenter[1] = oldmid[1];
	baricenter[2] = oldmid[2];

	xmin = oldmid[0]-(igrid-1)/(2*scale);
	ymin = oldmid[1]-(igrid-1)/(2*scale);
	zmin = oldmid[2]-(igrid-1)/(2*scale);

	xmax = oldmid[0]+(igrid-1)/(2*scale);
	ymax = oldmid[1]+(igrid-1)/(2*scale);
	zmax = oldmid[2]+(igrid-1)/(2*scale);

	cout << endl << INFO << "MAX " << xmax << " " << ymax << " " << zmax;
	cout << endl << INFO << "MIN " << xmin << " " << ymin << " "  << zmin;
	cout << endl << INFO << "Perfil " << perfill << " %";
	cout << endl << INFO << "Rmaxdim " << rmaxdim;

	cout << endl << INFO << "Allocating memory...";
	cout.flush();

	if (x!=NULL)
		delete x;
	if (y!=NULL)
		delete y;
	if (z!=NULL)
		delete z;

	x = allocateVector<double>(igrid);
	y = allocateVector<double>(igrid);
	z = allocateVector<double>(igrid);

	side = 1/scale;
	hside = side/2;
	A = side*side;

	if (buildStatus)
		if (status!=NULL)
			deleteMatrix3D<short>(nx,ny,status);

	nx = igrid;
	ny = igrid;
	nz = igrid;

	for (int i=0;i<nx;i++)
		x[i]=xmin+i*side;

	for (int i=0;i<ny;i++)
		y[i]=ymin+i*side;

	for (int i=0;i<nz;i++)
		z[i]=zmin+i*side;

	// allocate epsmap memory	
	bool flag = true;
	
	if (buildEpsMap)
		flag = clearAndAllocEpsMaps();

	// allocate grid point status memory
	// and initialize all to temporary outside status

	if (buildStatus)
	{
		status = allocateMatrix3D<short>(nx,ny,nz);
		for (int i=0;i<nx;i++)
			for (int j=0;j<ny;j++)
				for (int k=0;k<nz;k++)
					status[i][j][k]=STATUS_POINT_TEMPORARY_OUT;
	}

	if (buildEpsMap)
	{
		if (idebmap!=NULL)
			deleteVector<bool>(idebmap);

		idebmap = allocateVector<bool>(nx*ny*nz);
		// set to outside
		for (int i=0;i<nx;i++)
			for (int j=0;j<ny;j++)
				for (int k=0;k<nz;k++)
					IDEBMAP(i,j,k,nx,ny)=true;
	}

	if (!flag)
		return false;

	cout << "ok!";

	return true;

}

DelPhiShared::DelPhiShared(double scale,double perfill, char* fn,bool eps_flag,bool stat_flag)
{
	delphiBinding = false;
	buildEpsMap =eps_flag;
	buildStatus = stat_flag;
	cout << endl << INFO << "Initialization...";
	epsmap = NULL;
	idebmap = NULL;
	x = NULL;
	y = NULL;
	z = NULL;
	atoms = NULL;
	scspos = NULL;
	scsnor = NULL;
	scsarea = NULL;
	ibgp = NULL;
	status = NULL;
	cavitiesVec = NULL;

	cout << endl << INFO << "Loading atoms....";

	bool flag = loadAtoms(fn);

	if (!flag)
	{
		cout << endl << ERR << "Initialization failed";
		exit(-1);
	}

	flag = buildGrid(scale,perfill);
	if (!flag)
	{
		cout << endl << ERR << "Initialization failed";
		exit(-1);
	}
	cout << endl << INFO << "Initialization completed";
}

void DelPhiShared::DelPhiBinding(double xmin,double ymin,double zmin,
								 double xmax,double ymax,double zmax,
								 double c1,double c2,double c3,double rmax,double perf,
								 int* local_i_epsmap,int igrid,double scale,
							double* local_i_scspos,double* local_i_scsnor,
							bool* local_i_idebmap,int* local_i_ibgp,int maxbgp)
{
	cout << endl << INFO << "DelPhi Binding...";	
	delphiBinding = true;
	
	if (x!=NULL)
		delete x;
	if (y!=NULL)
		delete y;
	if (z!=NULL)
		delete z;

	this->xmin = xmin;
	this->ymin = ymin;
	this->zmin = zmin;

	this->rmaxdim=rmax;
	this->perfill = perf;
	this->xmax = xmax;
	this->ymax = ymax;
	this->zmax = zmax;

	this->baricenter[0]=c1;
	this->baricenter[1]=c2;
	this->baricenter[2]=c3;

	x = new double[igrid];
	y = new double[igrid];
	z = new double[igrid];

	this->scale = scale;
	side = 1/scale;
	hside = side/2;
	A = side*side;
	nx = igrid;
	ny = igrid;
	nz = igrid;

	atoms = NULL;
	
	// assuming all maps are already allocated before binding 
	this->epsmap  = local_i_epsmap;
	this->idebmap = local_i_idebmap;
	this->scsnor = local_i_scsnor;
	this->scsarea = NULL;
	this->scspos = local_i_scspos;
	this->ibgp = local_i_ibgp;

	for (int i=0;i<nx;i++)
		x[i]=xmin+i*side;

	for (int i=0;i<ny;i++)
		y[i]=ymin+i*side;

	for (int i=0;i<nz;i++)
		z[i]=zmin+i*side;

	// allocate grid point status memory
	// and initialize all to temporary outside status
	if (buildStatus)
	{
		status = allocateMatrix3D<short>(nx,ny,nz);
		for (int i=0;i<nx;i++)
			for (int j=0;j<ny;j++)
				for (int k=0;k<nz;k++)
					status[i][j][k]=2;
	}
	// set max bgp
	this->maxbgp = maxbgp;
	
	cout << "ok!";
	cout << endl << INFO << "Max number of bgps was set from DelPhi to " << maxbgp;
}


bool DelPhiShared::clearAndAllocEpsMaps()
{
	if (epsmap!=NULL)
		deleteVector<int>(epsmap);

	epsmap = allocateVector<int>(nx*ny*nz*3);

	if (epsmap == NULL)
	{
		cout << endl << ERR << "Error allocating epsmap!";
		return false;
	}
	
	for (int i=0;i<nx;i++)
			for (int j=0;j<ny;j++)
				for (int k=0;k<nz;k++)
				{
					EPSMAP(i,j,k,0,nx,ny,nz)=0;
					EPSMAP(i,j,k,1,nx,ny,nz)=0;
					EPSMAP(i,j,k,2,nx,ny,nz)=0;
				}
	return true;

}

void DelPhiShared::clearEpsMaps()
{

	if (epsmap == NULL)
	{
		cout << endl << ERR << "Cannot clear null map!";
		return;
	}
	
	for (int i=0;i<nx;i++)
			for (int j=0;j<ny;j++)
				for (int k=0;k<nz;k++)
				{
					EPSMAP(i,j,k,0,nx,ny,nz)=0;
					EPSMAP(i,j,k,1,nx,ny,nz)=0;
					EPSMAP(i,j,k,2,nx,ny,nz)=0;
				}
	return;
}

void DelPhiShared::saveEpsMaps(char* fname)
{
	if  (epsmap == NULL)
	{
		cout << endl << WARN << "Cannot save null epsmap!";
		return;
	}
	
	char f1[BUFLEN],f2[BUFLEN],f3[BUFLEN];
	sprintf(f1,"%s.epsmapx.txt",fname);
	sprintf(f2,"%s.epsmapy.txt",fname);
	sprintf(f3,"%s.epsmapz.txt",fname);

	FILE* fp = fopen(f1,"w");
	for (int i=0;i<nx;i++)
	{		
		for (int j=0;j<ny;j++)
		{
			for (int k=0;k<nz;k++)
				fprintf(fp,"%d ",EPSMAP(i,j,k,0,nx,ny,nz));
			fprintf(fp,"\n");
		}		
		fprintf(fp,"\n");
	}
	fclose(fp);

	fp = fopen(f2,"w");
	for (int i=0;i<nx;i++)
	{		
		for (int j=0;j<ny;j++)
		{
			for (int k=0;k<nz;k++)
				fprintf(fp,"%d ",EPSMAP(i,j,k,1,nx,ny,nz));
			fprintf(fp,"\n");
		}
		fprintf(fp,"\n");
	}		
	fclose(fp);

	fp = fopen(f3,"w");
	for (int i=0;i<nx;i++)
	{		
		for (int j=0;j<ny;j++)
		{
			for (int k=0;k<nz;k++)
				fprintf(fp,"%d ",EPSMAP(i,j,k,2,nx,ny,nz));
			fprintf(fp,"\n");
		}
		fprintf(fp,"\n");
	}		
	fclose(fp);
}

DelPhiShared::~DelPhiShared()
{
	if (epsmap!=NULL)
		deleteVector<int>(epsmap);
	if  (ibgp!=NULL)
		deleteVector<int>(ibgp);
	if  (scspos!=NULL)
		deleteVector<double>(scspos);
	if  (scsnor!=NULL)
		deleteVector<double>(scsnor);
	if  (scsarea!=NULL)
		deleteVector<double>(scsarea);
	if (x!=NULL)
		deleteVector<double>(x);
	if (y!=NULL)
		deleteVector<double>(y);
	if (z!=NULL)
		deleteVector<double>(z);	
	if (status!=NULL)
		deleteMatrix3D<short>(nx,ny,status);
	if (idebmap!=NULL)
		deleteVector<bool>(idebmap);

	if (atoms!=NULL)
	{
		for (int i=0;i<numAtoms;i++)
			delete atoms[i];
		delete[] atoms;
	}

	// free memory 
	if (cavitiesVec !=NULL)
	{
		vector< vector < int* >* >::iterator it;
		for (it=cavitiesVec->begin();it!=cavitiesVec->end();it++)
		{
			vector < int* >* inner = (*it);
			vector < int* >::iterator it2;
			for (it2=inner->begin();it2!=inner->end();it2++)
				free((*it2));
			delete inner;
		}		
	}
	delete cavitiesVec;
}

void DelPhiShared::saveBGP(char* fname)
{
	if  (scspos == NULL || scsnor==NULL)
	{
		cout << endl << WARN << "Cannot save null scspos or scsnor!";
		return;
	}

	char ff[BUFLEN];
	sprintf(ff,"%s.projections.txt",fname);		
	FILE* fp = fopen(ff,"w");
	FILE* fp2 = fopen("surf.txt","w");
	fprintf(fp,"%d\n",nbgp);
	for (int iibgp=0;iibgp<nbgp;iibgp++)
	{
		fprintf(fp,"%d %d %d %lf %lf %lf %lf %lf %lf",ibgp[iibgp*3]+1,ibgp[iibgp*3+1]+1,ibgp[iibgp*3+2]+1,
			scspos[iibgp*3],scspos[iibgp*3+1],scspos[iibgp*3+2],
			scsnor[iibgp*3],scsnor[iibgp*3+1],scsnor[iibgp*3+2]);

		fprintf(fp2,"%lf %lf %lf %lf %lf %lf",scspos[iibgp*3],scspos[iibgp*3+1],scspos[iibgp*3+2],
			scsnor[iibgp*3],scsnor[iibgp*3+1],scsnor[iibgp*3+2]);

		if (iibgp<nbgp-1)
		{
			fprintf(fp,"\n");
			fprintf(fp2,"\n");
		}
	}		
	fclose(fp);
	fclose(fp2);
}

void DelPhiShared::saveStatus(char* fname)
{
	if  (status == NULL)
	{
		cout << endl << WARN << "Cannot save null status!";
		return;
	}
	char ff[BUFLEN];
	sprintf(ff,"%s.status.txt",fname);
	FILE* fp = fopen(ff,"w");

	FILE* fp2 = fopen("cavities.txt","w");
	for (int i=0;i<nx;i++)
	{		
		for (int j=0;j<ny;j++)
		{
			for (int k=0;k<nz;k++)
			{
				fprintf(fp,"%d ",status[i][j][k]);
				if (status[i][j][k]>3)
					fprintf(fp2,"\n %f %f %f",x[i],y[j],z[k]);
			}
			fprintf(fp,"\n");
		}
		fprintf(fp,"\n");
	}
	fclose(fp);
	fclose(fp2);

	fp2 = fopen("cavitiesSize.txt","w");	
	vector <double>::iterator itCav; 	
	vector <int*>::iterator itCavVec;
	int i=0;

	fprintf(fp2,"%f\n",((float)cavitiesSize.size()));
	// for each cavity save the baricenter and the estimated volume
	for (itCav =  cavitiesSize.begin(),i=0; itCav!=cavitiesSize.end();itCav++,i++)
	{
		double barx,bary,barz;
		barx = 0;
		bary = 0;
		barz = 0;
		for (itCavVec =  cavitiesVec->at(i)->begin(); itCavVec!=cavitiesVec->at(i)->end();itCavVec++)
		{
			barx += x[(*itCavVec)[0]];
			bary += y[(*itCavVec)[1]];
			barz += z[(*itCavVec)[2]];
		}
		double volSize = cavitiesSize.at(i);
		int size = cavitiesVec->at(i)->size();
		barx/=size;
		bary/=size;
		barz/=size;
		int fillFlag;

		if (cavitiesFlag[i])
			fillFlag = 1;
		else
			fillFlag = 0;
		// save baricenter,volume, and filling flag
		fprintf(fp2,"%lf %lf %lf %lf %d\n",barx,bary,barz,volSize,fillFlag);	
	}
	fclose(fp2);		 
}