
/*
NanoShaper: a tool for geometric computations at the nano scale
Copyright (C) 2013  Sergio Decherchi

 This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "globals.h"
#include "MeshSurface.h"
#include "ExternalSurface.h"
#include "SkinSurface.h"
#include "BlobbySurface.h"
#include "ConnollySurface.h"
#include "tools.h"
#include "ConfigFile.h"

int num_cpus;

#ifndef DELPHI_BIND
/** @mainpage 
	@brief 
	The NanoShaper package is a framework to analyze an arbitrary surface, in particular a molecular surface.
	The following surfaces are provided: loading of a triangulated mesh (OFF or PLY file), loading
	of MSMS files, building the skin surface, the blobby surface and the Connolly surface.\n
	The Surface class is the main class of the framework: in order to expand the 
	framework with another surface type, the surface class must be extended and in particular
	some methods must be provided (see Surface class). The surface class is designed in a way such that
	if the extended class supports ray-surface intersection (projection is optional) then Surface class methods
	will do the rest of the work. Indeed such methods are sufficient to ray cast the surface and thus
	obtaining the in/out map from which a precise estimation of the volume, the triangulation, the total surface
	area can be computed; additionally cavity detection  and conditional volume filling can be performed. \n
	
	@author Sergio Decherchi
	@date 14-10-2012
	@version 0.3.1
	*/
int main(int argc,char* argv[])
{

	#ifdef  DBGMEM_CRT
		_CrtSetDbgFlag ( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
	#endif

	cout << endl << endl << INFO << "Starting " << PROGNAME << " version "<< VERSION ;
			
	ConfigFile* cf;
	char refName[BUFLEN];

	int numargs = argc-1;
	char confFile[BUFLEN];
	if (!numargs)
		strcpy(confFile,"surfaceConfiguration.prm");
	else
		strcpy(confFile,argv[1]);

	if (numargs>1)
		cout << endl << WARN << "Ignoring additional non required parameters";

	// get data from configuration file
	try
	{
		cf = new ConfigFile(confFile);
	}
	catch(...)
	{
		cout << endl << ERR << "Cannot read " << confFile;
		exit(-1);			
	}

	// mesh,msms,skin,external and grid data
	string mode = cf->read<string>( "Surface", "skin" );
	string fname = cf->read<string>( "Surface_File_Name", "mesh.off" );
	bool fillCavities = cf->read<bool>( "Cavity_Detection_Filling", false );
	bool saveEpsmaps = cf->read<bool>( "Save_eps_maps", false );
	bool buildEpsmaps = cf->read<bool>( "Build_epsilon_maps", false );
	bool buildStatus = cf->read<bool>( "Build_status_map", false );
	bool saveBgps = cf->read<bool>( "Save_bgps", false );
	bool saveStatusMap = cf->read<bool>( "Save_Status_map", false );
	double cavVol = cf->read<double>( "Conditional_Volume_Filling_Value", 11.4 );
	bool accTri = cf->read<bool>( "Accurate_Triangulation", false );
	bool tri = cf->read<bool>( "Triangulation", false );
	double skin_s = cf->read<double>( "Skin_Surface_Parameter", 0.45 );
	bool savePovRay = cf->read<bool>( "Save_PovRay", false );
	bool projBGP = cf->read<bool>( "Project_boundary_grid_points", false );
	double scale = cf->read<double>( "Grid_scale", 2.0 );
	double perfill = cf->read<double>( "Grid_perfil", 80.0 );
	string molFile = cf->read<string>( "XYZR_FileName");
	double blobby_B = cf->read<double>( "Blobbyness", -2.5 );
	bool checkDuplicatedVertices = cf->read<bool>( "Check_duplicated_vertices", true );
	
	unsigned int maxMeshDim = cf->read<unsigned int>( "Max_mesh_auxiliary_grid_size", 100 );
	unsigned int maxMeshPatches = cf->read<unsigned int>( "Max_mesh_patches_per_auxiliary_grid_cell", 250 );
	unsigned int maxMeshDim2D = cf->read<unsigned int>( "Max_mesh_auxiliary_grid_2d_size", 100 );
	unsigned int maxMeshPatches2D = cf->read<unsigned int>( "Max_mesh_patches_per_auxiliary_grid_2d_cell", 250 );
	
	unsigned int maxSESDim2D = cf->read<unsigned int>( "Max_ses_patches_auxiliary_grid_2d_size", 50 );
	unsigned int maxSESPatches2D = cf->read<unsigned int>( "Max_ses_patches_per_auxiliary_grid_2d_cell", 400 );
	unsigned int maxSESDim = cf->read<unsigned int>( "Max_ses_patches_auxiliary_grid_size", 100 );
	unsigned int maxSESPatches = cf->read<unsigned int>( "Max_ses_patches_per_auxiliary_grid_cell", 400 );
	
	unsigned int maxSkinDim = cf->read<unsigned int>( "Max_skin_patches_auxiliary_grid_size", 100 );
	unsigned int maxSkinPatches = cf->read<unsigned int>( "Max_skin_patches_per_auxiliary_grid_cell", 400 );
	unsigned int maxSkinDim2D = cf->read<unsigned int>( "Max_skin_patches_auxiliary_grid_2d_size", 50 );
	unsigned int maxSkinPatches2D = cf->read<unsigned int>( "Max_skin_patches_per_auxiliary_grid_2d_cell", 400 );
	
	int numMSMSFiles = cf->read<int>( "Num_MSMS_files", 1 );
	bool wellShaped = cf->read<bool>( "Keep_Water_Shaped_Cavities", false );
	double probeRadius = cf->read<double>( "Probe_Radius", 1.4 );
	bool smoothing = cf->read<bool>( "Smooth_Mesh", false );
	int numthd = cf->read<int>( "Number_thread", -1 );

	num_cpus = numthd;

	char mol[BUFLEN];

	if (!buildEpsmaps && saveEpsmaps)
	{
		cout << endl << ERR << "Asked to save epsmap without builiding it";
		cout << endl << REMARK << "Please set Build_epsilon_maps = true";
		cout << endl;
		exit(-1);
	}

	if (!buildEpsmaps && projBGP)
	{
		cout << endl << ERR << "Cannot project boundary grid points without an epsilon map.";
		cout << endl << REMARK << "Please set Build_epsilon_maps = true";
		cout << endl;
		exit(-1);	
	}

	if (!accTri && !buildStatus && tri)
	{
		// status map is needed to deduced in/out vertices
		buildStatus = true;
		cout << endl << ERR << "If non analytical triangulation is enabled status map is needed.";
		cout << endl << REMARK << "Please set Build_status_map = true";
		cout << endl;
		exit(-1);
	}

	if (fillCavities && !buildStatus)
	{
		// status map is needed to search cavities
		buildStatus = true;	
		cout << endl << ERR << "If cavity detection is enabled status map is needed.";
		cout << endl << REMARK << "Please set Build_status_map = true";
		cout << endl;
		exit(-1);
	}

	strcpy(mol,molFile.c_str());
	// Set up DelPhi-like environment
	DelPhiShared* dg = new DelPhiShared(scale,perfill,mol,buildEpsmaps,buildStatus);
	
	Surface* surf;
	bool load_f;

	// MeshSurface
	if (!mode.compare("mesh") || !mode.compare("msms"))
	{
		// Bind DelphiShared with surface engine
		MeshSurface* msurf = new MeshSurface(dg);	
		msurf->setAuxGrid(maxMeshDim,maxMeshPatches);
		msurf->setAuxGrid2D(maxMeshDim2D,maxMeshPatches2D);
		// Set up inside value
		msurf->inside = 5;
		
		if (!mode.compare("mesh"))
		{
			// Load Surface
			load_f = msurf->load((char*)fname.c_str());
			if (!load_f)
			{
				cout << endl << ERR << "Cannot load " << fname;
				exit(-1);
			}
		}
		else
		{
			// Load MSMS surface[s]
			load_f = ((MeshSurface*)msurf)->loadMSMS((char*)fname.c_str(),numMSMSFiles);
			if (!load_f)
			{
				cout << endl << ERR << "Cannot load " << fname;
				exit(-1);
			}
			char cc[BUFLEN];
			strcpy(cc,"msms.off");
			msurf->save(cc);
		}
		surf = msurf;
		strcpy(refName,fname.c_str());
	}

	// External Surface
	else if (!mode.compare("external"))
	{
		// Bind DelphiShared with surface engine
		ExternalSurface* esurf = new ExternalSurface(dg);	
		// Link delphi code
		esurf->inside = 5;
		surf = esurf;
		strcpy(refName,"external");
	}
	
	// Skin Surface
	else if (!mode.compare("skin"))
	{
		SkinSurface* ssurf = new SkinSurface(dg);
		ssurf->setAuxGrid(maxSkinDim,maxSkinPatches);
		ssurf->setAuxGrid2D(maxSkinDim2D,maxSkinPatches2D);
		ssurf->inside = 5;
		surf = ssurf;
		ssurf->setShrinking(skin_s);
		strcpy(refName,"skin");
		ssurf->setSavePovRay(savePovRay);
	}
	// Connolly Surface
	else if (!mode.compare("ses"))
	{
		ConnollySurface* csurf = new ConnollySurface(dg);
		csurf->setAuxGrid(maxSESDim,maxSESPatches);
		csurf->setAuxGrid2D(maxSESDim2D,maxSESPatches2D);
		csurf->inside = 5;
		surf = csurf;
		csurf->setProbeRadius(probeRadius);
		strcpy(refName,"ses");
		csurf->setSavePovRay(savePovRay);
	}

	// Blobby Surface
	else if (!mode.compare("blobby"))
	{
		BlobbySurface* bsurf = new BlobbySurface(dg);
		bsurf->setAuxGrid(maxMeshDim,maxMeshPatches);
		bsurf->setAuxGrid2D(maxMeshDim2D,maxMeshPatches2D);
		bsurf->inside = 5;
		surf = bsurf;
		bsurf->setBlobbyness(blobby_B);		
		strcpy(refName,"blobby");		
	}
	else
	{
		cout << endl << ERR << "Unknown surface";
		exit(-1);
	}
	
	surf->setProjBGP(projBGP);
	surf->setTriangulationFlag(accTri);
	surf->setCheckDuplicatedVertices(checkDuplicatedVertices);
	surf->setKeepWellShapedCavities(wellShaped);
	surf->setProbeRadius(probeRadius);
	surf->setSmoothing(smoothing);
	
	time_t start, end;
	double duration = 0;
	time(&start);	
	// Pre-process surface
	surf->build();

	// Build DelPhi stuff
	surf->getSurf(fillCavities,cavVol);		
	time(&end);
	duration = difftime(end, start);
	cout << endl << INFO << "Surface computation time.. " << duration << " [s]"; 	
	cout << endl << INFO << "Estimated volume by averaged triple ray-casting " << surf->getVolume() << " [A^3]"; 
		
	if (saveEpsmaps)
	{
		cout << endl << INFO << "Saving epsmaps...";
		// Save epsmap
		dg->saveEpsMaps(refName);
		cout << "ok!";
	}

	if (saveBgps)
	{
		cout << endl << INFO << "Saving bgpmap...";
		dg->saveBGP(refName);
		cout << "ok!";
	}

	if (saveStatusMap)
	{
		cout << endl << INFO << "Saving statusmap...";
		dg->saveStatus(refName);
		cout << "ok!";	
	}
	
	if (tri)
	{
		if (accTri)
			cout << endl << INFO << "Accurate triangulation enabled";

		cout << endl << INFO << "Triangulating Surface...";				
		time(&start);
		surf->triangulateSurface();		
		if (surf->getSmoothing())
		{
			cout << endl << INFO << "Smoothing surface...";
			surf->smoothSurface();			
		}
		time(&end);
		duration = difftime(end, start);
		cout << "ok!";
		cout << endl << INFO << "Total Triangulation time " << duration << " [s]";
	}
				
	delete surf;
	delete dg;
	delete cf;
		
	// Memory leak detection 
	#ifdef DBGMEM_CRT
		_CrtDumpMemoryLeaks();
	#endif

	cout << endl << INFO << "Closing " << PROGNAME;
	cout << endl << endl ;
	
	return 0;
	
}

#endif

	#ifdef DELPHI_BIND
	
	/** This is the DelPhi Fortran-C binding. Returns the number of bgp */
	// check visual, borland c++, xlc, watcom, for windows
	#if  (defined _WIN32) || (defined __WIN32__) || (defined __TOS_WIN__) || (defined __WINDOWS__) 
			extern "C" __declspec( dllexport )		
	// linux,unix,apple,android 
	#elif (defined __linux) || (defined __unix) || (defined macintosh) || (defined Macintosh) || (defined __APPLE__ && defined __MACH__) || (defined __ANDROID__)
			extern "C" 		
	#endif
					
	void epsmakemodule(double xmin,double ymin,double zmin,
					  double xmax,double ymax,double zmax,
					  double c1,double c2,double c3,	double rmax,double perf,							 
					  int* i_epsmap,int igrid,double scale,double* i_scspos,
				      double* i_scsnor,bool* i_idebmap,int* i_ibgp,int inside,int* ibnum,int natom,
					  double* xn1,double* rad, int* d,int maxbgp,double probe)
	{
	  
		cout << endl << endl << INFO << "Starting " << PROGNAME << " " << VERSION;
		

		ConfigFile* cf;
		// get data from default configuration file
		try
		{
			cf = new ConfigFile( "surfaceConfiguration.prm" );
		}
		catch(...)
		{
			cout << endl << ERR << "Cannot read surfaceConfiguration.prm";
			cout.flush();
			exit(-1);			
		}		

		// mesh,msms,skin,external
		string mode = cf->read<string>( "Surface", "ses" );
		string fname = cf->read<string>( "Surface_File_Name", "triangulatedSurf.off" );
		bool fillCavities = cf->read<bool>( "Cavity_Detection_Filling", false );
		bool saveEpsmaps = cf->read<bool>( "Save_eps_maps", false );
		bool saveBgps = cf->read<bool>( "Save_bgps", false );
		bool buildEpsmaps = cf->read<bool>("Build_epsilon_maps", false);
		bool buildStatus = cf->read<bool>("Build_status_map", false);
		bool saveStatusMap = cf->read<bool>( "Save_Status_map", false );
		double cavVol = cf->read<double>( "Conditional_Volume_Filling_Value", 11.4 );
		bool accTri = cf->read<bool>( "Accurate_Triangulation", false );
		bool tri = cf->read<bool>( "Triangulation", false );
		double skin_s = cf->read<double>( "Skin_Surface_Parameter", 0.45 );
		bool savePovRay = cf->read<bool>( "Save_PovRay", false );
		bool projBGP = cf->read<bool>( "Project_boundary_grid_points", true );				
		double blobby_B = cf->read<double>( "Blobbyness", -2.5 );
		bool checkDuplicatedVertices = cf->read<bool>( "Check_duplicated_vertices", true );
		int maxTri = cf->read<int>( "Max_triangles_per_auxiliary_grid_cell", 250 );
		int numMSMSFiles = cf->read<int>( "Num_MSMS_files", 1 );
		bool wellShaped = cf->read<bool>( "Keep_Water_Shaped_Cavities", false );
		//double probeRadius = cf->read<double>( "Probe_Radius", 1.4 );
		double probeRadius = probe;
		bool smoothing = cf->read<bool>( "Smooth_Mesh", false );
		int numthd = cf->read<int>( "Number_thread", 32 );

		unsigned int maxMeshDim = cf->read<unsigned int>( "Max_mesh_auxiliary_grid_size", 100 );
		unsigned int maxMeshPatches = cf->read<unsigned int>( "Max_mesh_patches_per_auxiliary_grid_cell", 250 );
		unsigned int maxMeshDim2D = cf->read<unsigned int>( "Max_mesh_auxiliary_grid_2d_size", 100 );
		unsigned int maxMeshPatches2D = cf->read<unsigned int>( "Max_mesh_patches_per_auxiliary_grid_2d_cell", 250 );
	
		unsigned int maxSESDim2D = cf->read<unsigned int>( "Max_ses_patches_auxiliary_grid_2d_size", 50 );
		unsigned int maxSESPatches2D = cf->read<unsigned int>( "Max_ses_patches_per_auxiliary_grid_2d_cell", 400 );
		unsigned int maxSESDim = cf->read<unsigned int>( "Max_ses_patches_auxiliary_grid_size", 100 );
		unsigned int maxSESPatches = cf->read<unsigned int>( "Max_ses_patches_per_auxiliary_grid_cell", 400 );
	
		unsigned int maxSkinDim = cf->read<unsigned int>( "Max_skin_patches_auxiliary_grid_size", 100 );
		unsigned int maxSkinPatches = cf->read<unsigned int>( "Max_skin_patches_per_auxiliary_grid_cell", 400 );
		unsigned int maxSkinDim2D = cf->read<unsigned int>( "Max_skin_patches_auxiliary_grid_2d_size", 50 );
		unsigned int maxSkinPatches2D = cf->read<unsigned int>( "Max_skin_patches_per_auxiliary_grid_2d_cell", 400 );

		delete cf;
		
		// get data from custom configuration file
		try
		{
			cf = new ConfigFile( "custom.prm" );
		}
		catch(...)
		{
			cout << endl << ERR << "Cannot read custom.prm";
			cout.flush();
			exit(-1);			
		}		
		mode = cf->read<string>( "Surface", "ses" );
		fname = cf->read<string>( "Surface_File_Name", "triangulatedSurf.off" );

		num_cpus = numthd;

		if (!accTri && !buildStatus && tri)
		{
			// status map is needed to deduced in/out vertices
			buildStatus = true;
			cout << endl << ERR << "If non analytical triangulation is enabled status map is needed. Please set Build_status_map = true";
			cout << endl;
			exit(-1);
		}

		if (fillCavities && !buildStatus)
		{
			// status map is needed to search cavities
			buildStatus = true;	
			cout << endl << ERR << "If cavity detection is enabled status map is needed. Please set Build_status_map = true";
			cout << endl;
			exit(-1);
		}

		char refName[BUFLEN];
		
		cout << endl << INFO << "Binding with DelPhi..";		
		// Set up delphi environment
		DelPhiShared* dg = new DelPhiShared();
		dg->DelPhiBinding(	xmin,ymin,zmin,xmax,ymax,zmax,
							c1,c2,c3,rmax,perf,
							i_epsmap,igrid,scale,
							i_scspos,i_scsnor,i_idebmap,i_ibgp,maxbgp);

		// load atoms info if present
		dg->loadAtoms(natom,xn1,rad,NULL,NULL);
		// here only means populate epsilon map
		dg->buildEpsmap(true);

		cout << "ok!";
		cout << endl << INFO << "DelPhi grid is " << igrid;
		Surface* surf;
		bool load_f;

		// MeshSurface
		if (!mode.compare("mesh") || !mode.compare("msms"))
		{
			// Bind DelphiShared with surface engine
			MeshSurface* msurf = new MeshSurface(dg);	
			// Set up inside value
			msurf->inside = (inside);
			msurf->setAuxGrid(maxMeshDim,maxMeshPatches);
			msurf->setAuxGrid2D(maxMeshDim2D,maxMeshPatches2D);
			
			if (!mode.compare("mesh"))
			{
				// Load Surface
				load_f = msurf->load((char*)fname.c_str());
				if (!load_f)
				{
					cout << endl << ERR << "Cannot load " << fname;
					exit(-1);
				}
			}
			else
			{
				// Load MSMS surface[s]
				load_f = ((MeshSurface*)msurf)->loadMSMS((char*)fname.c_str(),numMSMSFiles);
				if (!load_f)
				{
					cout << endl << ERR << "Cannot load " << fname;
					exit(-1);
				}
				char cc[BUFLEN];
				strcpy(cc,"msms.off");
				msurf->save(cc);
			}
			surf = msurf;			
			strcpy(refName,fname.c_str());
		}

		// External Surface
		else if (!mode.compare("external"))
		{
			// Bind DelphiShared with surface engine
			ExternalSurface* esurf = new ExternalSurface(dg);	
			// Link delphi code
			esurf->inside = (inside);
			surf = esurf;
			strcpy(refName,"external");
		}
		
		// Skin Surface
		else if (!mode.compare("skin"))
		{
			SkinSurface* ssurf = new SkinSurface(dg);
			ssurf->inside = (inside);
			ssurf->setAuxGrid(maxSkinDim,maxSkinPatches);
			ssurf->setAuxGrid2D(maxSkinDim2D,maxSkinPatches2D);
			surf = ssurf;
			ssurf->setShrinking(skin_s);
			strcpy(refName,"skin");
			ssurf->setSavePovRay(savePovRay);
		}
		// Connolly Surface
		else if (!mode.compare("ses"))
		{
			ConnollySurface* csurf = new ConnollySurface(dg);
			csurf->inside = (inside);
			csurf->setAuxGrid(maxSESDim,maxSESPatches);
			csurf->setAuxGrid2D(maxSESDim2D,maxSESPatches2D);
			surf = csurf;
			csurf->setProbeRadius(probeRadius);
			strcpy(refName,"ses");
			csurf->setSavePovRay(savePovRay);
		}
		// Blobby Surface
		else if (!mode.compare("blobby"))
		{
			BlobbySurface* bsurf = new BlobbySurface(dg);
			bsurf->inside = (inside);
			bsurf->setAuxGrid(maxMeshDim,maxMeshPatches);
			bsurf->setAuxGrid2D(maxMeshDim2D,maxMeshPatches2D);
			surf = bsurf;
			bsurf->setBlobbyness(blobby_B);
			strcpy(refName,"blobby");		
		}
		else
		{
			cout << endl << ERR << "Unknown surface";
			exit(-1);
		}
		
		surf->setProjBGP(projBGP);
		surf->setTriangulationFlag(accTri);
		surf->setCheckDuplicatedVertices(checkDuplicatedVertices);
		surf->setKeepWellShapedCavities(wellShaped);
		surf->setProbeRadius(probeRadius);
		surf->setSmoothing(smoothing);
		
		time_t start, end;
		double duration = 0;
		time(&start);
		
		// Pre-process surface
		surf->build();

		// Build DelPhi stuff
		surf->getSurf(fillCavities,cavVol);		
		time(&end);
		duration = difftime(end, start);
		cout << endl << INFO << "Surface computation time.. " << duration << " [s]"; 	
		cout << endl << INFO << "Estimated volume by averaged triple ray-casting " << surf->getVolume() << " [A^3]"; 
			
		if (saveEpsmaps)
		{
			cout << endl << INFO << "Saving epsmaps...";
			// Save epsmap
			dg->saveEpsMaps(refName);
			cout << "ok!";
		}

		if (saveBgps)
		{
			cout << endl << INFO << "Saving bgpmap...";
			dg->saveBGP(refName);
			cout << "ok!";
		}

		if (saveStatusMap)
		{
			cout << endl << INFO << "Saving statusmap...";
			dg->saveStatus(refName);
			cout << "ok!";	
		}
		
		if (tri)
		{
			if (accTri)
				cout << endl << INFO << "Accurate triangulation enabled";

			cout << endl << INFO << "Triangulating Surface...";				
			time(&start);
			surf->triangulateSurface();		
			if (surf->getSmoothing())
			{
				cout << endl << INFO << "Smoothing surface...";
				surf->smoothSurface();			
			}
			time(&end);
			duration = difftime(end, start);
			cout << "ok!";
			cout << endl << INFO << "Triangulation time " << duration;
		}
					
		// avoid to destroy things that now belong to DelPhi
		dg->finalizeBinding();		
		// save number of boundary grid points
		(*ibnum) = dg->nbgp;
		
		delete surf;
		delete dg;
		delete cf;
		
		cout << endl << INFO << "End " << PROGNAME << " Module";
		cout << endl << endl ;
		return;
	}
	
	#endif