
//---------------------------------------------------------
/*    @file		tools.h
*     @brief	tools.h Includes some common tools
*							    							*/
//---------------------------------------------------------

#ifndef tools_h
#define tools_h

#include "jama_eig.h"
#include "globals.h"
#include "./sturm/solve.h"
#include <string>

using namespace TNT;
using namespace JAMA;
// Miscellanous utility functions and lightweighted classes
#define MAX_POLY_DEGREE 10 // max polynomial degree supported by the bgp projection routine (6 is needed for skin, 4 by Connolly)

/**@brief A point in R^3*/
class Point
{
public:
	double pos[3];	
	Point()
	{
		pos[0]=0;
		pos[1]=0;
		pos[2]=0; 
	}
	Point(double x,double y,double z)
	{
		pos[0] = x;
		pos[1] = y;
		pos[2] = z;
	}
	Point(Point& p)
	{
		pos[0]= p.pos[0];
		pos[1]= p.pos[1];
		pos[2]= p.pos[2];
	}
	
	double squaredDist(Point* x)
	{
		double acc =0;
		for (int i=0;i<3;i++)
		{	
			double t = x->pos[i]-pos[i];
			acc += t*t;
		}
		return acc;
	}

	double dot(Point* x)
	{
		double acc =0;
		for (int i=0;i<3;i++)
			acc += x->pos[i]*pos[i];
		return acc;
	}


	double norm(Point* x)
	{
		double acc =0;
		for (int i=0;i<3;i++)
			acc += (pos[i])*(pos[i]);
		return sqrt(acc);
	}
	
	double dist(Point* x)
	{
		double acc = 0;
		for (int i=0;i<3;i++)
		{
			double t = x->pos[i]-pos[i];
			acc += t*t;
		}
		return sqrt(acc);
	}

	void sumInPlace(Point* x)
	{		
		for (int i=0;i<3;i++)
		{
			pos[0]+=x->pos[0];	
			pos[1]+=x->pos[1];
			pos[2]+=x->pos[2];
		}
	}

	Point* sum(Point* x)
	{	
		Point* p = new Point();
		for (int i=0;i<3;i++)
		{
			p->pos[0]=pos[0]+x->pos[0];	
			p->pos[1]=pos[1]+x->pos[1];	
			p->pos[2]=pos[2]+x->pos[2];	
		}
		return p;
	}
};

/**@brief An atom: an atom is a point in R^3 with radius, charge and dielectric value*/
class Atom: public Point
{
public:
	double radius;
	double charge;
	int dielectric;

	Atom():Point()
	{
		radius=0;
		charge=0;
		dielectric=0; 
	}
	
	Atom(double x,double y,double z,double r,double c,int d):Point(x,y,z)
	{
		radius = r;
		charge = c;
		dielectric = d;
	}
	
	// copy constructor
	Atom(Atom& p):Point(p)
	{
		radius = p.radius;
		charge = p.charge;
		dielectric = p.dielectric;
	}

};


////////////////// Memory management routines //////////////////////

template<class T> T* allocateVector(int n)
{
	T* t = (T*)malloc(sizeof(T)*n);
	if (t==NULL)
	{
		cout << endl << ERR << "Not enough memory to allocate vector ";
		return NULL;
	}
	return t;
}

template<class T>  T** allocateMatrix2D(int nrows,int ncol)
{
	T** t = (T**)malloc(sizeof(T*)*nrows);
	if (t==NULL)
	{
		cout << endl << ERR << "Not enough memory to allocate 2D matrix ";
		return NULL;
	}
	for (int i=0;i<nrows;i++)
	{
		t[i] = (T*)malloc(sizeof(T)*ncol);
		
		if (t[i]==NULL)
		{
			cout << endl << ERR << "Not enough memory to allocate 2D matrix ";
			return NULL;
		}	
	}
	return t;
}

template<class T>  T*** allocateMatrix3D(int nx,int ny,int nz)
{
	T*** t = (T***)malloc(sizeof(T**)*nx);
	if (t==NULL)
	{
		cout << endl << ERR << "Not enough memory to allocate 3D matrix ";
		return NULL;
	}
	for (int i=0;i<nx;i++)
	{
		t[i] = (T**)malloc(sizeof(T*)*ny);
		if (t[i]==NULL)
		{	
			cout << endl << ERR << "Not enough memory to allocate 3D matrix ";
			return NULL;
		}

		for (int j=0;j<ny;j++)
		{
			t[i][j] =  (T*)malloc(sizeof(T)*nz);
			if (t[i][j]==NULL)
			{	
				cout << endl << ERR << "Not enough memory to allocate 3D matrix ";
				return NULL;
			}
		}
	}
	return t;
}

template<class T> void deleteVector(T* t)
{
	free(t);
}

template<class T> void deleteMatrix2D(int nrows,T** t)
{
	for (int i=0;i<nrows;i++)
		free(t[i]);
	free(t);
}

template<class T> void deleteMatrix3D(int nx,int ny,T*** t)
{
	for (int i=0;i<nx;i++)
	{
		for (int j=0;j<ny;j++)
			free(t[i][j]);
		free(t[i]);
	}
	free(t);
}
///////////////////////////////////////////////////////////////////


///////////////////// STL comparators //////////////////////////////
/**@brief ascending on first double of pair<double,double*> comparator*/
bool compKeepIndex(pair<double,double*> a, pair<double,double*> b);
///////////////////////////////////////////////////////////////////

typedef std::pair<double,int> indexed_double;
bool index_double_comparator( const indexed_double& l, const indexed_double& r);
///////////////////// MISCELLANEAOUS ////////////////////////////////

/** @brief test if point proj is in cube whose center is point and side is side variable.
 If a toll is provided, the test is performed up to a tollerance value.
 By default the tollerance is 0*/
bool testInCube(double proj0,double proj1,double proj2,double point0,double point1, double point2,double side,double toll=0);
double rintp(double x);
string toLowerCase(string str);
void cleanLine();

static long int SEED=1234;
// randnumber between 0 and 1
double randnum();

/** get the real roots by computing the companion matrix and then extracting the eigenvalues. A root is real
if its imaginary part in absolute value is less than a given threshold. Usually this threshold is rather conservative
such that possibly unprecise real roots are not lost*/
void getRealRootsCompanion(double* poly,int degree,double* roots,int& numroots);
/** get real roots by using Sturm method. Directly search real roots*/
void getRealRootsSturm(double* poly,int degree,double* roots,int& numroots);
void plane3points(double p1[3],double p2[3],double p3[3],double w[4],bool normalize=true);
void point2plane(double p[3], double w[4],double* dist, double proj[3]);
void inplace_invert4x4(double M[4][4]);
void Matrix4x4MultiplyBy4x4 (double src1[4][4], double src2[4][4], double dest[4][4]);

static	int shift_map[27][3]=	{	{0,0,0},
						{0,0,-1},
						{0,0,+1},
						{0,+1,0},
						{0,+1,-1},
						{0,+1,+1},
						{0,-1,0},
						{0,-1,-1},
						{0,-1,+1},

						{+1,0,0},
						{+1,0,-1},
						{+1,0,+1},
						{+1,+1,0},
						{+1,+1,-1},
						{+1,+1,+1},
						{+1,-1,0},
						{+1,-1,-1},
						{+1,-1,+1},

						{-1,0,0},
						{-1,0,-1},
						{-1,0,+1},
						{-1,+1,0},
						{-1,+1,-1},
						{-1,+1,+1},
						{-1,-1,0},
						{-1,-1,-1},
						{-1,-1,+1}	};
#endif